import json
import logging
from json import JSONDecodeError
from typing import Dict, List, Union

from ruamel.yaml import CommentedMap, CommentedSeq, YAML
from ruamel.yaml.composer import ComposerError


def read_json(path: str) -> List[Dict]:
    with open(path) as f:
        try:
            decoded = json.load(f)
        except JSONDecodeError:
            logging.error(f'Failed to parse {path}. JSON invalid')
            exit(1)
        else:
            if type(decoded) == dict:
                decoded = [decoded]

            return decoded


def load_yaml(yaml_str: str, path: str) -> Union[CommentedMap, CommentedSeq]:
    yaml = YAML()
    yaml.preserve_quotes = True
    try:
        yaml_file = yaml.load(yaml_str)
    except ComposerError:
        logging.error(f'Failed to parse {path}. YAML invalid')
        exit(1)
    else:
        return yaml_file


def read_yaml(path: str) -> Union[List[Dict], List[CommentedMap]]:
    decoded = []

    # 1. read the file
    with open(path) as f:
        # strip manifest divisor if it is in the beginning or in the end
        raw = f.readlines()

        # ignore comments
        raw = ''.join([i for i in raw if not i.startswith('#')])

    # 2. Check if items are separated by "---"
    if "---" in raw:
        raw = raw.strip('---').split('---')
        for r in raw:
            y = load_yaml(r, path)
            decoded.append(y)

        return decoded

    # 3. The file should be readable as is by yaml safe load
    yaml_decoded = load_yaml(raw, path)
    # return list of dicts
    if type(yaml_decoded) == CommentedMap:
        decoded = [yaml_decoded]
    elif type(yaml_decoded) == CommentedSeq:
        for cm in yaml_decoded:
            decoded.append(cm)

    return decoded



import os
from copy import deepcopy

import pytest
from pydantic import ValidationError

from src.ksux.src.patch import read_patches, get_real_path, apply_patch, validate_patch
from src.ksux.src.schemas import Patch, Op, Action
from src.ksux.tests.config import settings


def get_valid_patch() -> dict:
    return {
        "name": "valid patch",
        "target": {
            "apiVersion": "apps/v1",
            "kind": "Deployment",
            "name": "deploy"
        },
        "ops": [
            {
                "name": "example op",
                "path": "/spec/replicas",
                "action": "replace",
                "value": "2",
                "enforce_integer": False
            }
        ]
    }


def test_validate_patch_valid_patch():
    valid = get_valid_patch()
    assert Patch(**valid)

    # Test enforce integer False
    patch = get_valid_patch()
    patch["ops"][0]["enforce_integer"] = False
    validated = validate_patch(patch)
    assert isinstance(validated.ops[0].value, str)

    # Test enforce integer True
    patch["ops"][0]["enforce_integer"] = True
    validated = validate_patch(patch)
    assert isinstance(validated.ops[0].value, int)

    # value can be omitted for action remove
    valid_patch_action_remove_missing_value = get_valid_patch()
    valid_patch_action_remove_missing_value['ops'][0]['action'] = 'remove'
    del valid_patch_action_remove_missing_value['ops'][0]['value']

    assert Patch(**valid_patch_action_remove_missing_value)


def test_validate_patch_missing_field():
    # All fields must be present
    invalid_patch_missing_target_name = get_valid_patch()
    del invalid_patch_missing_target_name['target']['name']

    with pytest.raises(ValidationError) as exc:
        Patch(**invalid_patch_missing_target_name)

    # Value must be present for action replace or add
    invalid_patch_missing_value_for_replace_add = get_valid_patch()
    del invalid_patch_missing_value_for_replace_add['ops'][0]['value']

    for a in ['replace', 'add']:
        invalid_patch_missing_value_for_replace_add['ops'][0]['action'] = a

        with pytest.raises(ValueError) as exc:
            Patch(**invalid_patch_missing_value_for_replace_add)


def test_validate_patch_add_check_type():
    patch_add = get_valid_patch()
    patch_add['ops'][0]['action'] = 'add'
    patch_add['ops'][0]['value'] = {'key': 'val'}

    assert Patch(**patch_add)

    invalid_patch_value_for_add_not_object = get_valid_patch()
    invalid_patch_value_for_add_not_object['ops'][0]['action'] = 'add'
    invalid_patch_value_for_add_not_object['ops'][0]['value'] = 'val'

    with pytest.raises(ValueError):
        Patch(**invalid_patch_value_for_add_not_object)


def test_read_patches():
    p = read_patches(os.path.join(settings.PARENT, 'data', 'patches'))

    assert type(p) == list
    assert len(p) > 0


def test_get_real_path():
    manifest = {
        "apiVersion": "apps/v1",
        "kind": "Deployment",
        "metadata": {
            "name": "deploy"
        },
        "spec": {
            "template": {
                "spec": {
                    "containers": [
                        {
                            "name": "container1",
                            "image": "nginx"
                        },
                        {
                            "name": "container2",
                            "image": "redis"
                        },
                        {
                            "name": "container3",
                            "image": "mariadb"
                        },
                        {
                            "name": "container4",
                            "image": "python"
                        },
                    ]
                }
            }
        }
    }

    patch = Patch(**get_valid_patch())

    for i in range(1, 5):
        patch.ops[0].path = f'/spec/template/spec/containers/container{i}/image'
        grp = get_real_path(patch, patch.ops[0], manifest)

        assert grp == ['spec', 'template', 'spec', 'containers', i - 1, 'image']


def test_patch_manifest():
    manifest = {
        "apiVersion": "apps/v1",
        "kind": "Deployment",
        "metadata": {
            "name": "deploy"
        },
        "spec": {
            "template": {
                "spec": {
                    "replicas": 1,
                    "containers": [
                        {
                            "name": "container1",
                            "image": "nginx"
                        }
                    ]
                }
            }
        }
    }

    replicas = '5'
    new_image_name = 'new-image-name'
    new_image = 'redis'

    patched_manifest_real = deepcopy(manifest)
    patched_manifest_real['spec']['template']['spec']['replicas'] = replicas
    patched_manifest_real['spec']['template']['spec']['containers'][0]['name'] = new_image_name
    patched_manifest_real['spec']['template']['spec']['containers'][0]['image'] = new_image

    patch = Patch(**get_valid_patch())
    patch.ops = [
        Op(name='replicas', path='/spec/template/spec/replicas', action=Action.replace, value=replicas),
        Op(name='image name', path='/spec/template/spec/containers/container1/name', action=Action.replace,
           value=new_image_name),
        Op(name='image', path=f'/spec/template/spec/containers/{new_image_name}/image', action=Action.replace,
           value=new_image)
    ]

    patched_manifest = apply_patch(patch, patch.ops[0], manifest)
    patched_manifest = apply_patch(patch, patch.ops[1], patched_manifest)
    patched_manifest = apply_patch(patch, patch.ops[2], patched_manifest)


def test_patch_ingress():
    """Patch Ingress resource

    Just an example, since ingress contains host field instead of name
    """

    NEW_HOST = 'new_host'

    manifest = {
        "apiVersion": "networking.k8s.io/v1",
        "kind": "Ingress",
        "metadata": {
            "name": "ingress"
        },
        "spec": {
            "rules": [
                {
                    "host": "host1",
                    "http": {
                        "paths": [
                            {
                                "path": "/",
                                "pathType": "Prefix",
                                "backend": {
                                    "service": {
                                        "name": "svc",
                                        "port": {
                                            "number": 80
                                        }
                                    }
                                }
                            }
                        ]
                    }
                }
            ]
        }
    }

    patch_dict = {
        "name": "ingress patch",
        "target": {
            "apiVersion": "networking.k8s.io/v1",
            "kind": "Ingress",
            "name": "ingress"
        },
        "ops": [
            {
                "name": "replace host",
                "path": "/spec/rules/host1/host",
                "action": "replace",
                "value": NEW_HOST,
                "list_key": "host"
            }
        ]
    }
    patch = Patch(**patch_dict)

    patched = apply_patch(patch, patch.ops[0], manifest)

    assert patched['spec']['rules'][0]['host'] == NEW_HOST



import logging
import os
from typing import Dict, List, Tuple

from .patch import read_patches, patch_manifest
from .read_file import read_yaml, read_json

logging.basicConfig(level=logging.DEBUG)


def add_manifest(manifests_index: dict, manifest: dict) -> None:
    """
    Add manifest to manifests_index
    :param manifests_index: index of manifests
    :param manifest: manifest to be added
    """
    try:
        apiVersion = manifest['apiVersion']
        kind = manifest['kind']
        name = manifest['metadata']['name']
    except KeyError:
        logging.error('apiVersion, kind and/or metadata.name key is not present')
        exit(1)
    else:
        if apiVersion not in manifests_index:
            manifests_index[apiVersion] = {}

        if kind not in manifests_index[apiVersion]:
            manifests_index[apiVersion][kind] = {}

        if name in manifests_index[apiVersion][kind]:
            logging.error('There are two manifests with the same apiVersion, kind and name. This is forbidden')
            exit(1)
        else:
            manifests_index[apiVersion][kind][name] = manifest


def read_manifests(base_dir: str, exclude: List[str] = []) -> Dict:
    """
    Read manifests from the base dir and create a manifest index
    :param base_dir: directory with manifest templates
    :param exclude: exclude pattern ["{apiVersion}_{kind}_{name}", ...]
    :return: manifest index
    """
    manifests = {}
    for m in os.listdir(base_dir):
        file_path = os.path.join(base_dir, m)
        if m.endswith('yaml') or m.endswith('yml'):
            temp = read_yaml(file_path)
        elif m.endswith('.json'):
            temp = read_json(file_path)
        else:
            logging.info(f'Unknown extension ".{m.split(".")[-1]}". Skipping...')
            continue

        for manifest in temp:
            if f'{manifest["apiVersion"]}_{manifest["kind"]}_{manifest["metadata"]["name"]}' not in exclude:
                add_manifest(manifests, manifest)

    return manifests


def patch_manifests(base_dir: str, patches_dir: str, exclude: List[str]) -> Tuple[List[Dict], Dict]:
    """
    Patch all manifests
    :param base_dir: directory with manifest templates
    :param patches_dir: patches directory
    :param exclude: exclude pattern ["{apiVersion}_{kind}_{name}", ...]
    :return: list of patched manifests + all manifests
    """
    manifests = read_manifests(base_dir=base_dir, exclude=exclude)
    patches = read_patches(patches_dir=patches_dir, exclude=exclude)

    patched = []
    for patch in patches:
        patched.append(patch_manifest(patch, manifests))

    return patched, manifests



import json
import logging
import os.path
from typing import List, Dict

from ruamel.yaml import round_trip_dump


def save_manifests(
        patched_manifests: List[Dict],
        out_dir: str,
        extension: str
) -> None:
    os.system(f'mkdir -p {out_dir}')

    for p in patched_manifests:
        output_path = os.path.join(
            out_dir,
            f'{p["apiVersion"].replace("/", "-")}_{p["kind"]}_{p["metadata"]["name"]}.{extension}'.lower()
        )

        with open(output_path, 'w') as f:
            if extension == 'json':
                json.dump(p, f, indent=2, ensure_ascii=False)
            elif extension == 'yaml' or extension == 'yml':
                round_trip_dump(p, f)
            else:
                logging.error('Invalid extension. Allowed values are: json, yaml and yml')
                exit(1)



import json
import logging
import sys
from argparse import ArgumentParser

from ruamel.yaml import round_trip_dump

from .src.create_manifests import patch_manifests
from .src.save_manifests import save_manifests


def main():
    parser = ArgumentParser()

    parser.add_argument('-b', '--base_dir', help='Base directory with yaml/yml/json kubernetes templates')
    parser.add_argument('-o', '--out_dir', help='Output directory', default='out')
    parser.add_argument('-p', '--patches_dir',
                        help='Patches directory. Script will read only files with yaml/yml/json extensions')
    parser.add_argument('-e', '--output_extension', help='Extension of output files', choices=['yaml', 'json', 'yml'],
                        default='json')
    parser.add_argument('--patched_only', help='If all manifests should be saved, even those which were not patched',
                        action='store_true')
    parser.add_argument('-x', '--exclude', help='Exclude one or more manifests. To target the manifest, argument '
                                                'expects format: {apiVersion}_{kind}_{name}. E.g. if you wished to '
                                                'exclude deployment with name "backend", the option would look like: '
                                                'ksux -b ... -p ... -o ... -x apps/v1_Deployment_backend',
                        action='append')
    parser.add_argument('-xf', '--exclude-file', help='File with a list of resources to be excluded. Resources'
                                                      'must follow the format: {apiVersion}_{kind}_{name}')
    parser.add_argument('--dry-run', help='Print manifests to stdout', action="store_true")
    parser.add_argument('-q', '--quiet', help='Do not print debug, info and warning messages', action='store_true')
    parser.add_argument('--version', help='Package version', action="store_true")

    args = parser.parse_args()

    if args.version:
        print('0.6.0')
        exit(0)

    if args.quiet:
        logging.root.setLevel(logging.ERROR)
    else:
        logging.root.setLevel(logging.INFO)

    exclude = [] if not args.exclude else args.exclude

    if args.exclude_file:
        try:
            with open(args.exclude_file) as f:
                for line in f.readlines():
                    res = line.strip()

                    # ignore comments (lines starting with "#") and empty lines
                    if res and not res.startswith('#'):
                        exclude.append(res)
        except FileNotFoundError:
            logging.error(f'File {args.exclude_file} not found')

    # validate exclude
    for e in exclude:
        try:
            apiVersion, kind, name = e.split('_')
        except ValueError:
            logging.error('Exclude items have to be in format: {apiVersion}_{kind}_{name}')
            exit(1)

    final, all_manifests = patch_manifests(base_dir=args.base_dir, patches_dir=args.patches_dir, exclude=exclude)

    if args.dry_run:
        logging.info('Showing patched manifests.')
        if args.output_extension in ['yaml', 'yml']:
            print('---')
            for i, m in enumerate(final):
                round_trip_dump(m, sys.stdout)
                print('---')
        else:
            print(json.dumps(final))
    else:
        # Create final list of manifests
        if not args.patched_only:
            versions = all_manifests.keys()
            final = []

            for v in versions:
                version = all_manifests[v]
                kinds = version.keys()
                for k in kinds:
                    kind = version[k]
                    resource_names = kind.keys()

                    for r in resource_names:
                        final.append(kind[r])

        save_manifests(patched_manifests=final, out_dir=args.out_dir, extension=args.output_extension)

    if args.output_extension in ['yaml', 'yml']:
        logging.warning(
            'Make sure to check the manifests, especially if your input was in json format and you had some "add" ops '
            'in your patches. In this case it is impossible to preserve quotes correctly!')
